@@ -1,16 +1,24 @@
 from random import Random
 from lru import LRU
 
-from eth2spec.phase0 import spec as spec_phase0
-from eth2spec.test.helpers.attestations import cached_prepare_state_with_attestations
-from eth2spec.test.helpers.deposits import mock_deposit
-from eth2spec.test.helpers.state import next_epoch
+from eth2spec.phase0.mainnet import VALIDATOR_REGISTRY_LIMIT  # equal everywhere, fine to import
+from eth2spec.test.context import is_post_altair
+from eth2spec.test.helpers.state import (
+    next_epoch,
+)
+from eth2spec.test.helpers.random import (
+    set_some_new_deposits, exit_random_validators, slash_random_validators,
+    randomize_state,
+)
+from eth2spec.test.helpers.attestations import (
+    cached_prepare_state_with_attestations,
+)
 from eth2spec.utils.ssz.ssz_typing import Container, uint64, List
 
 
 class Deltas(Container):
-    rewards: List[uint64, spec_phase0.VALIDATOR_REGISTRY_LIMIT]
-    penalties: List[uint64, spec_phase0.VALIDATOR_REGISTRY_LIMIT]
+    rewards: List[uint64, VALIDATOR_REGISTRY_LIMIT]
+    penalties: List[uint64, VALIDATOR_REGISTRY_LIMIT]
 
 
 def has_enough_for_reward(spec, state, index):
@@ -26,6 +34,26 @@ def has_enough_for_reward(spec, state, index):
     )
 
 
+def has_enough_for_leak_penalty(spec, state, index):
+    """
+    Check if effective_balance and state of leak is high enough for a leak penalty.
+
+    At very low balances / leak values, it is possible for a validator have a positive effective_balance
+    and be in a leak, but have zero leak penalty.
+    """
+
+    if is_post_altair(spec):
+        return (
+            state.validators[index].effective_balance * state.inactivity_scores[index]
+            > spec.config.INACTIVITY_SCORE_BIAS * spec.INACTIVITY_PENALTY_QUOTIENT_ALTAIR
+        )
+    else:
+        return (
+            state.validators[index].effective_balance * spec.get_finality_delay(state)
+            > spec.INACTIVITY_PENALTY_QUOTIENT
+        )
+
+
 def run_deltas(spec, state):
     """
     Run all deltas functions yielding:
@@ -33,35 +61,58 @@ def run_deltas(spec, state):
       - source deltas ('source_deltas')
       - target deltas ('target_deltas')
       - head deltas ('head_deltas')
-      - inclusion delay deltas ('inclusion_delay_deltas')
+      - not if is_post_altair(spec)
+          - inclusion delay deltas ('inclusion_delay_deltas')
       - inactivity penalty deltas ('inactivity_penalty_deltas')
     """
     yield 'pre', state
+
+    if is_post_altair(spec):
+        def get_source_deltas(state):
+            return spec.get_flag_index_deltas(state, spec.TIMELY_SOURCE_FLAG_INDEX)
+
+        def get_head_deltas(state):
+            return spec.get_flag_index_deltas(state, spec.TIMELY_HEAD_FLAG_INDEX)
+
+        def get_target_deltas(state):
+            return spec.get_flag_index_deltas(state, spec.TIMELY_TARGET_FLAG_INDEX)
+
     yield from run_attestation_component_deltas(
         spec,
         state,
-        spec.get_source_deltas,
+        spec.get_source_deltas if not is_post_altair(spec) else get_source_deltas,
         spec.get_matching_source_attestations,
         'source_deltas',
     )
     yield from run_attestation_component_deltas(
         spec,
         state,
-        spec.get_target_deltas,
+        spec.get_target_deltas if not is_post_altair(spec) else get_target_deltas,
         spec.get_matching_target_attestations,
         'target_deltas',
     )
     yield from run_attestation_component_deltas(
         spec,
         state,
-        spec.get_head_deltas,
+        spec.get_head_deltas if not is_post_altair(spec) else get_head_deltas,
         spec.get_matching_head_attestations,
         'head_deltas',
     )
-    yield from run_get_inclusion_delay_deltas(spec, state)
+    if not is_post_altair(spec):
+        yield from run_get_inclusion_delay_deltas(spec, state)
     yield from run_get_inactivity_penalty_deltas(spec, state)
 
 
+def deltas_name_to_flag_index(spec, deltas_name):
+    if 'source' in deltas_name:
+        return spec.TIMELY_SOURCE_FLAG_INDEX
+    elif 'head' in deltas_name:
+        return spec.TIMELY_HEAD_FLAG_INDEX
+    elif 'target' in deltas_name:
+        return spec.TIMELY_TARGET_FLAG_INDEX
+    raise ValueError("Wrong deltas_name %s" % deltas_name)
+
+
 def run_attestation_component_deltas(spec, state, component_delta_fn, matching_att_fn, deltas_name):
     """
     Run ``component_delta_fn``, yielding:
@@ -71,8 +122,14 @@ def run_attestation_component_deltas(spec, state, component_delta_fn, matching_a
 
     yield deltas_name, Deltas(rewards=rewards, penalties=penalties)
 
-    matching_attestations = matching_att_fn(state, spec.get_previous_epoch(state))
-    matching_indices = spec.get_unslashed_attesting_indices(state, matching_attestations)
+    if not is_post_altair(spec):
+        matching_attestations = matching_att_fn(state, spec.get_previous_epoch(state))
+        matching_indices = spec.get_unslashed_attesting_indices(state, matching_attestations)
+    else:
+        matching_indices = spec.get_unslashed_participating_indices(
+            state, deltas_name_to_flag_index(spec, deltas_name), spec.get_previous_epoch(state)
+        )
+
     eligible_indices = spec.get_eligible_validator_indices(state)
     for index in range(len(state.validators)):
         if index not in eligible_indices:
@@ -83,14 +140,23 @@ def run_attestation_component_deltas(spec, state, component_delta_fn, matching_a
         validator = state.validators[index]
         enough_for_reward = has_enough_for_reward(spec, state, index)
         if index in matching_indices and not validator.slashed:
-            if enough_for_reward:
-                assert rewards[index] > 0
+            if is_post_altair(spec):
+                if not spec.is_in_inactivity_leak(state) and enough_for_reward:
+                    assert rewards[index] > 0
+                else:
+                    assert rewards[index] == 0
             else:
-                assert rewards[index] == 0
+                if enough_for_reward:
+                    assert rewards[index] > 0
+                else:
+                    assert rewards[index] == 0
+
             assert penalties[index] == 0
         else:
             assert rewards[index] == 0
-            if enough_for_reward:
+            if is_post_altair(spec) and 'head' in deltas_name:
+                assert penalties[index] == 0
+            elif enough_for_reward:
                 assert penalties[index] > 0
             else:
                 assert penalties[index] == 0
@@ -101,6 +167,12 @@ def run_get_inclusion_delay_deltas(spec, state):
     Run ``get_inclusion_delay_deltas``, yielding:
       - inclusion delay deltas ('inclusion_delay_deltas')
     """
+    if is_post_altair(spec):
+        # No inclusion_delay_deltas
+        yield 'inclusion_delay_deltas', Deltas(rewards=[0] * len(state.validators),
+                                               penalties=[0] * len(state.validators))
+        return
+
     rewards, penalties = spec.get_inclusion_delay_deltas(state)
 
     yield 'inclusion_delay_deltas', Deltas(rewards=rewards, penalties=penalties)
@@ -148,8 +220,13 @@ def run_get_inactivity_penalty_deltas(spec, state):
 
     yield 'inactivity_penalty_deltas', Deltas(rewards=rewards, penalties=penalties)
 
-    matching_attestations = spec.get_matching_target_attestations(state, spec.get_previous_epoch(state))
-    matching_attesting_indices = spec.get_unslashed_attesting_indices(state, matching_attestations)
+    if not is_post_altair(spec):
+        matching_attestations = spec.get_matching_target_attestations(state, spec.get_previous_epoch(state))
+        matching_attesting_indices = spec.get_unslashed_attesting_indices(state, matching_attestations)
+    else:
+        matching_attesting_indices = spec.get_unslashed_participating_indices(
+            state, spec.TIMELY_TARGET_FLAG_INDEX, spec.get_previous_epoch(state)
+        )
 
     eligible_indices = spec.get_eligible_validator_indices(state)
     for index in range(len(state.validators)):
@@ -159,32 +236,55 @@ def run_get_inactivity_penalty_deltas(spec, state):
             continue
 
         if spec.is_in_inactivity_leak(state):
+            # Compute base_penalty
             base_reward = spec.get_base_reward(state, index)
-            base_penalty = spec.BASE_REWARDS_PER_EPOCH * base_reward - spec.get_proposer_reward(state, index)
+            if not is_post_altair(spec):
+                cancel_base_rewards_per_epoch = spec.BASE_REWARDS_PER_EPOCH
+                base_penalty = cancel_base_rewards_per_epoch * base_reward - spec.get_proposer_reward(state, index)
+
             if not has_enough_for_reward(spec, state, index):
                 assert penalties[index] == 0
-            elif index in matching_attesting_indices:
-                assert penalties[index] == base_penalty
+            elif index in matching_attesting_indices or not has_enough_for_leak_penalty(spec, state, index):
+                if is_post_altair(spec):
+                    assert penalties[index] == 0
+                else:
+                    assert penalties[index] == base_penalty
             else:
-                assert penalties[index] > base_penalty
+                if is_post_altair(spec):
+                    assert penalties[index] > 0
+                else:
+                    assert penalties[index] > base_penalty
         else:
-            assert penalties[index] == 0
+            if not is_post_altair(spec):
+                assert penalties[index] == 0
+                continue
+            else:
+                # post altair, this penalty is derived from the inactivity score
+                # regardless if the state is leaking or not...
+                if index in matching_attesting_indices:
+                    assert penalties[index] == 0
+                else:
+                    # copied from spec:
+                    penalty_numerator = state.validators[index].effective_balance * state.inactivity_scores[index]
+                    penalty_denominator = spec.config.INACTIVITY_SCORE_BIAS * spec.INACTIVITY_PENALTY_QUOTIENT_ALTAIR
+                    assert penalties[index] == penalty_numerator // penalty_denominator
 
 
 def transition_state_to_leak(spec, state, epochs=None):
     if epochs is None:
-        epochs = spec.MIN_EPOCHS_TO_INACTIVITY_PENALTY
-    assert epochs >= spec.MIN_EPOCHS_TO_INACTIVITY_PENALTY
+        # +2 because finality delay is based on previous_epoch and must be more than `MIN_EPOCHS_TO_INACTIVITY_PENALTY`
+        epochs = spec.MIN_EPOCHS_TO_INACTIVITY_PENALTY + 2
+    assert epochs > spec.MIN_EPOCHS_TO_INACTIVITY_PENALTY
 
     for _ in range(epochs):
         next_epoch(spec, state)
+    assert spec.is_in_inactivity_leak(state)
 
 
 _cache_dict = LRU(size=10)
 
 
 def leaking(epochs=None):
-
     def deco(fn):
         def entry(*args, spec, state, **kw):
             # If the pre-state is not already known in the LRU, then take it,
@@ -204,49 +304,6 @@ def leaking(epochs=None):
     return deco
 
 
-def set_some_new_deposits(spec, state, rng):
-    num_validators = len(state.validators)
-    # Set ~1/10 to just recently deposited
-    for index in range(num_validators):
-        # If not already active, skip
-        if not spec.is_active_validator(state.validators[index], spec.get_current_epoch(state)):
-            continue
-        if rng.randrange(num_validators) < num_validators // 10:
-            mock_deposit(spec, state, index)
-            # Set ~half of selected to eligible for activation
-            if rng.choice([True, False]):
-                state.validators[index].activation_eligibility_epoch = spec.get_current_epoch(state)
-
-
-def exit_random_validators(spec, state, rng):
-    if spec.get_current_epoch(state) < 5:
-        # Move epochs forward to allow for some validators already exited/withdrawable
-        for _ in range(5):
-            next_epoch(spec, state)
-
-    current_epoch = spec.get_current_epoch(state)
-    # Exit ~1/2 of validators
-    for index in spec.get_active_validator_indices(state, current_epoch):
-        if rng.choice([True, False]):
-            continue
-
-        validator = state.validators[index]
-        validator.exit_epoch = rng.choice([current_epoch - 1, current_epoch - 2, current_epoch - 3])
-        # ~1/2 are withdrawable
-        if rng.choice([True, False]):
-            validator.withdrawable_epoch = current_epoch
-        else:
-            validator.withdrawable_epoch = current_epoch + 1
-
-
-def slash_random_validators(spec, state, rng):
-    # Slash ~1/2 of validators
-    for index in range(len(state.validators)):
-        # slash at least one validator
-        if index == 0 or rng.choice([True, False]):
-            spec.slash_validator(state, index)
-
-
 def run_test_empty(spec, state):
     # Do not add any attestations to state
 
@@ -262,8 +319,13 @@ def run_test_full_all_correct(spec, state):
 def run_test_full_but_partial_participation(spec, state, rng=Random(5522)):
     cached_prepare_state_with_attestations(spec, state)
 
-    for a in state.previous_epoch_attestations:
-        a.aggregation_bits = [rng.choice([True, False]) for _ in a.aggregation_bits]
+    if not is_post_altair(spec):
+        for a in state.previous_epoch_attestations:
+            a.aggregation_bits = [rng.choice([True, False]) for _ in a.aggregation_bits]
+    else:
+        for index in range(len(state.validators)):
+            if rng.choice([True, False]):
+                state.previous_epoch_participation[index] = spec.ParticipationFlags(0b0000_0000)
 
     yield from run_deltas(spec, state)
 
@@ -272,8 +334,12 @@ def run_test_partial(spec, state, fraction_filled):
     cached_prepare_state_with_attestations(spec, state)
 
     # Remove portion of attestations
-    num_attestations = int(len(state.previous_epoch_attestations) * fraction_filled)
-    state.previous_epoch_attestations = state.previous_epoch_attestations[:num_attestations]
+    if not is_post_altair(spec):
+        num_attestations = int(len(state.previous_epoch_attestations) * fraction_filled)
+        state.previous_epoch_attestations = state.previous_epoch_attestations[:num_attestations]
+    else:
+        for index in range(int(len(state.validators) * fraction_filled)):
+            state.previous_epoch_participation[index] = spec.ParticipationFlags(0b0000_0000)
 
     yield from run_deltas(spec, state)
 
@@ -328,13 +394,18 @@ def run_test_some_very_low_effective_balances_that_attested(spec, state):
 def run_test_some_very_low_effective_balances_that_did_not_attest(spec, state):
     cached_prepare_state_with_attestations(spec, state)
 
-    # Remove attestation
-    attestation = state.previous_epoch_attestations[0]
-    state.previous_epoch_attestations = state.previous_epoch_attestations[1:]
-    # Set removed indices effective balance to very low amount
-    indices = spec.get_unslashed_attesting_indices(state, [attestation])
-    for i, index in enumerate(indices):
-        state.validators[index].effective_balance = i
+    if not is_post_altair(spec):
+        # Remove attestation
+        attestation = state.previous_epoch_attestations[0]
+        state.previous_epoch_attestations = state.previous_epoch_attestations[1:]
+        # Set removed indices effective balance to very low amount
+        indices = spec.get_unslashed_attesting_indices(state, [attestation])
+        for i, index in enumerate(indices):
+            state.validators[index].effective_balance = i
+    else:
+        index = 0
+        state.validators[index].effective_balance = 1
+        state.previous_epoch_participation[index] = spec.ParticipationFlags(0b0000_0000)
 
     yield from run_deltas(spec, state)
 
@@ -436,22 +507,5 @@ def run_test_all_balances_too_low_for_reward(spec, state):
 
 
 def run_test_full_random(spec, state, rng=Random(8020)):
-    set_some_new_deposits(spec, state, rng)
-    exit_random_validators(spec, state, rng)
-    slash_random_validators(spec, state, rng)
-
-    cached_prepare_state_with_attestations(spec, state)
-
-    for pending_attestation in state.previous_epoch_attestations:
-        # ~1/3 have bad target
-        if rng.randint(0, 2) == 0:
-            pending_attestation.data.target.root = b'\x55' * 32
-        # ~1/3 have bad head
-        if rng.randint(0, 2) == 0:
-            pending_attestation.data.beacon_block_root = b'\x66' * 32
-        # ~50% participation
-        pending_attestation.aggregation_bits = [rng.choice([True, False]) for _ in pending_attestation.aggregation_bits]
-        # Random inclusion delay
-        pending_attestation.inclusion_delay = rng.randint(1, spec.SLOTS_PER_EPOCH)
-
+    randomize_state(spec, state, rng)
     yield from run_deltas(spec, state)
