@@ -1,7 +1,11 @@
-from eth2spec.test.context import spec_state_test, with_all_phases
-from eth2spec.test.phase0.epoch_processing.run_epoch_process_base import (
+from random import Random
+from eth2spec.test.context import spec_state_test, with_all_phases, is_post_altair
+from eth2spec.test.helpers.epoch_processing import (
     run_epoch_processing_with, run_epoch_processing_to
 )
+from eth2spec.test.helpers.random import randomize_state
+from eth2spec.test.helpers.state import has_active_balance_differential
+from eth2spec.test.helpers.voluntary_exits import get_unslashed_exited_validators
 from eth2spec.test.helpers.state import next_epoch
 
 
@@ -22,25 +26,42 @@ def slash_validators(spec, state, indices, out_epochs):
         spec.get_current_epoch(state) % spec.EPOCHS_PER_SLASHINGS_VECTOR
     ] = total_slashed_balance
 
+    # verify some slashings happened...
+    assert total_slashed_balance != 0
 
-@with_all_phases
-@spec_state_test
-def test_max_penalties(spec, state):
+
+def get_slashing_multiplier(spec):
+    if is_post_altair(spec):
+        return spec.PROPORTIONAL_SLASHING_MULTIPLIER_ALTAIR
+    else:
+        return spec.PROPORTIONAL_SLASHING_MULTIPLIER
+
+
+def _setup_process_slashings_test(spec, state, not_slashable_set=set()):
     # Slashed count to ensure that enough validators are slashed to induce maximum penalties
     slashed_count = min(
-        (len(state.validators) // spec.PROPORTIONAL_SLASHING_MULTIPLIER) + 1,
+        (len(state.validators) // get_slashing_multiplier(spec)) + 1,
         # Can't slash more than validator count!
         len(state.validators)
     )
     out_epoch = spec.get_current_epoch(state) + (spec.EPOCHS_PER_SLASHINGS_VECTOR // 2)
 
-    slashed_indices = list(range(slashed_count))
-    slash_validators(spec, state, slashed_indices, [out_epoch] * slashed_count)
+    eligible_indices = set(range(slashed_count))
+    slashed_indices = eligible_indices.difference(not_slashable_set)
+    slash_validators(spec, state, sorted(slashed_indices), [out_epoch] * slashed_count)
 
     total_balance = spec.get_total_active_balance(state)
     total_penalties = sum(state.slashings)
 
-    assert total_balance // spec.PROPORTIONAL_SLASHING_MULTIPLIER <= total_penalties
+    assert total_balance // get_slashing_multiplier(spec) <= total_penalties
+
+    return slashed_indices
+
+
+@with_all_phases
+@spec_state_test
+def test_max_penalties(spec, state):
+    slashed_indices = _setup_process_slashings_test(spec, state)
 
     yield from run_process_slashings(spec, state)
 
@@ -50,9 +71,32 @@ def test_max_penalties(spec, state):
 
 @with_all_phases
 @spec_state_test
-def test_small_penalty(spec, state):
+def test_low_penalty(spec, state):
+    # Slashed count is one tenth of validator set
+    slashed_count = (len(state.validators) // 10) + 1
+    out_epoch = spec.get_current_epoch(state) + (spec.EPOCHS_PER_SLASHINGS_VECTOR // 2)
+
+    slashed_indices = list(range(slashed_count))
+    slash_validators(spec, state, slashed_indices, [out_epoch] * slashed_count)
+
+    pre_state = state.copy()
+
+    yield from run_process_slashings(spec, state)
+
+    for i in slashed_indices:
+        assert 0 < state.balances[i] < pre_state.balances[i]
+
+
+@with_all_phases
+@spec_state_test
+def test_minimal_penalty(spec, state):
+    #
+    # When very few slashings, the resulting slashing penalty gets rounded down
+    # to zero so the result of `process_slashings` is null
+    #
+
     # Just the bare minimum for this one validator
-    state.balances[0] = state.validators[0].effective_balance = spec.EJECTION_BALANCE
+    state.balances[0] = state.validators[0].effective_balance = spec.config.EJECTION_BALANCE
     # All the other validators get the maximum.
     for i in range(1, len(state.validators)):
         state.validators[i].effective_balance = state.balances[i] = spec.MAX_EFFECTIVE_BALANCE
@@ -74,11 +118,13 @@ def test_small_penalty(spec, state):
 
     expected_penalty = (
         state.validators[0].effective_balance // spec.EFFECTIVE_BALANCE_INCREMENT
-        * (3 * total_penalties)
+        * (get_slashing_multiplier(spec) * total_penalties)
         // total_balance
         * spec.EFFECTIVE_BALANCE_INCREMENT
     )
-    assert state.balances[0] == pre_slash_balances[0] - expected_penalty
+
+    assert expected_penalty == 0
+    assert state.balances[0] == pre_slash_balances[0]
 
 
 @with_all_phases
@@ -88,7 +134,7 @@ def test_scaled_penalties(spec, state):
     next_epoch(spec, state)
 
     # Also mock some previous slashings, so that we test to have the delta in the penalties computation.
-    base = spec.EJECTION_BALANCE
+    base = spec.config.EJECTION_BALANCE
     incr = spec.EFFECTIVE_BALANCE_INCREMENT
     # Just add some random slashings. non-zero slashings are at least the minimal effective balance.
     state.slashings[0] = base + (incr * 12)
@@ -96,7 +142,7 @@ def test_scaled_penalties(spec, state):
     state.slashings[5] = base + (incr * 6)
     state.slashings[spec.EPOCHS_PER_SLASHINGS_VECTOR - 1] = base + (incr * 7)
 
-    slashed_count = len(state.validators) // (spec.PROPORTIONAL_SLASHING_MULTIPLIER + 1)
+    slashed_count = len(state.validators) // (get_slashing_multiplier(spec) + 1)
 
     assert slashed_count > 10
 
@@ -134,8 +180,33 @@ def test_scaled_penalties(spec, state):
         v = state.validators[i]
         expected_penalty = (
             v.effective_balance // spec.EFFECTIVE_BALANCE_INCREMENT
-            * (spec.PROPORTIONAL_SLASHING_MULTIPLIER * total_penalties)
+            * (get_slashing_multiplier(spec) * total_penalties)
             // (total_balance)
             * spec.EFFECTIVE_BALANCE_INCREMENT
         )
         assert state.balances[i] == pre_slash_balances[i] - expected_penalty
+
+
+@with_all_phases
+@spec_state_test
+def test_slashings_with_random_state(spec, state):
+    rng = Random(9998)
+    randomize_state(spec, state, rng)
+
+    pre_balances = state.balances.copy()
+
+    target_validators = get_unslashed_exited_validators(spec, state)
+    assert len(target_validators) != 0
+    assert has_active_balance_differential(spec, state)
+
+    slashed_indices = _setup_process_slashings_test(spec, state, not_slashable_set=target_validators)
+
+    # ensure no accidental slashings of protected set...
+    current_target_validators = get_unslashed_exited_validators(spec, state)
+    assert len(current_target_validators) != 0
+    assert current_target_validators == target_validators
+
+    yield from run_process_slashings(spec, state)
+
+    for i in slashed_indices:
+        assert state.balances[i] < pre_balances[i]
